package net.knarcraft.knarlib.formatting;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import static net.knarcraft.knarlib.formatting.StringFormatter.replacePlaceholder;

/**
 * A utility for formatting a string specifying an amount of time
 */
@SuppressWarnings("unused")
public final class TimeFormatter {

    private static Map<Double, TranslatableTimeUnit[]> timeUnits;
    private static List<Double> sortedUnits;

    private TimeFormatter() {

    }

    /**
     * Gets the string used for displaying this sign's duration
     *
     * @param translator <p>The translator to use for translation</p>
     * @param duration   <p>The duration, in seconds, to display</p>
     * @return <p>The string used for displaying this sign's duration</p>
     */
    public static String getDurationString(Translator translator, long duration) {
        if (duration == 0) {
            return translator.getTranslatedMessage(TranslatableTimeUnit.UNIT_NOW);
        } else {
            if (sortedUnits == null) {
                initializeUnits(translator);
            }
            for (double unit : sortedUnits) {
                if (duration / unit >= 1) {
                    double units = round(duration / unit);
                    return formatDurationString(units, translator, timeUnits.get(unit)[units == 1 ? 0 : 1],
                            (units * 10) % 10 == 0);
                }
            }
            return formatDurationString(duration, translator, TranslatableTimeUnit.UNIT_SECONDS, false);
        }
    }

    /**
     * Rounds a number to its last two digits
     *
     * @param number <p>The number to round</p>
     * @return <p>The rounded number</p>
     */
    private static double round(double number) {
        return Math.round(number * 100.0) / 100.0;
    }

    /**
     * Formats a duration string
     *
     * @param duration            <p>The duration to display</p>
     * @param translator          <p>The translator to use for translation</p>
     * @param translatableMessage <p>The time unit to display</p>
     * @param castToInt           <p>Whether to cast the duration to an int</p>
     * @return <p>The formatted duration string</p>
     */
    private static String formatDurationString(double duration, Translator translator,
                                               TranslatableTimeUnit translatableMessage, boolean castToInt) {
        String durationFormat = translator.getTranslatedMessage(TranslatableTimeUnit.DURATION_FORMAT);
        durationFormat = replacePlaceholder(durationFormat, "{unit}",
                translator.getTranslatedMessage(translatableMessage));
        return replacePlaceholder(durationFormat, "{time}", castToInt ? String.valueOf((int) duration) :
                String.valueOf(duration));
    }

    /**
     * Initializes the mapping of available time units for formatting permission sign duration
     */
    private static void initializeUnits(Translator translator) {
        double minute = 60;
        double hour = 60 * minute;
        double day = 24 * hour;
        double week = 7 * day;
        double month = 4 * week;
        double year = 365 * day;
        double decade = 10 * year;

        timeUnits = new HashMap<>();
        registerTimeUnit(translator, decade, TranslatableTimeUnit.UNIT_DECADE, TranslatableTimeUnit.UNIT_DECADES);
        registerTimeUnit(translator, year, TranslatableTimeUnit.UNIT_YEAR, TranslatableTimeUnit.UNIT_YEARS);
        registerTimeUnit(translator, month, TranslatableTimeUnit.UNIT_MONTH, TranslatableTimeUnit.UNIT_MONTHS);
        registerTimeUnit(translator, week, TranslatableTimeUnit.UNIT_WEEK, TranslatableTimeUnit.UNIT_WEEKS);
        registerTimeUnit(translator, day, TranslatableTimeUnit.UNIT_DAY, TranslatableTimeUnit.UNIT_DAYS);
        registerTimeUnit(translator, hour, TranslatableTimeUnit.UNIT_HOUR, TranslatableTimeUnit.UNIT_HOURS);
        registerTimeUnit(translator, minute, TranslatableTimeUnit.UNIT_MINUTE, TranslatableTimeUnit.UNIT_MINUTES);
        if (!registerTimeUnit(translator, 1D, TranslatableTimeUnit.UNIT_SECOND, TranslatableTimeUnit.UNIT_SECONDS)) {
            throw new IllegalStateException("No translation has been registered for UNIT_SECOND and UNIT_SECONDS");
        }

        sortedUnits = new ArrayList<>(timeUnits.keySet());
        Collections.sort(sortedUnits);
        Collections.reverse(sortedUnits);
    }

    /**
     * Registers one time unit
     *
     * @param translator <p>The translator to use for checking whether necessary translations have been set</p>
     * @param seconds    <p>The amount of seconds the time unit corresponds to</p>
     * @param singular   <p>The translatable time unit to use for the singular form</p>
     * @param plural     <p>The translatable time unit to use for the plural form</p>
     */
    private static boolean registerTimeUnit(Translator translator, double seconds, TranslatableTimeUnit singular,
                                            TranslatableTimeUnit plural) {
        String singularTranslation = translator.getTranslatedMessage(singular);
        String pluralTranslation = translator.getTranslatedMessage(plural);
        if (singularTranslation != null && !singularTranslation.isBlank() && pluralTranslation != null &&
                !pluralTranslation.isBlank()) {
            timeUnits.put(seconds, new TranslatableTimeUnit[]{singular, plural});
            return true;
        } else {
            return false;
        }
    }

}
